# Copyright 2019 PIQuIL - All Rights Reserved.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import abc
import numpy as np

from .utils import _update_statistics


class ObservableBase(abc.ABC):
    """Base class for observables."""

    _name = None
    _symbol = None

    @property
    def name(self):
        """The name of the Observable."""
        if self._name is None:
            self._name = self.__class__.__name__
        return self._name

    @name.setter
    def name(self, new_name):
        self._name = new_name

    @property
    def symbol(self):
        """The algebraic symbol representing the Observable."""
        if self._symbol is None:
            self._symbol = self.__class__.__name__
        return self._symbol

    @symbol.setter
    def symbol(self, new_symbol):
        self._symbol = new_symbol

    def __str__(self):
        return self.symbol

    def __repr__(self):
        return self.name

    def __neg__(self):
        return ProdObservable(
            self, -1, name=("-" + self.name), symbol=("-" + self.symbol)
        )

    def __add__(self, other):
        return SumObservable(self, other)

    def __sub__(self, other):
        return SumObservable(self, -other)

    def __mul__(self, other):
        return ProdObservable(self, other)

    def __radd__(self, other):
        return SumObservable(other, self)

    def __rsub__(self, other):
        return SumObservable(other, -self)

    def __rmul__(self, other):
        return ProdObservable(other, self)

    @abc.abstractmethod
    def apply(self, nn_state, samples):
        """Computes the value of the observable, row-wise, on a batch of
        samples.

        If we think of the samples given as a set of projective measurements
        in a given computational basis, this method must return the expectation
        of the operator with respect to each basis state in `samples`.
        It must not perform any averaging for statistical purposes, as the
        proper analysis is delegated to the specialized
        `statistics` and `statistics_from_samples` methods.

        Must be implemented by any subclasses.

        :param nn_state: The WaveFunction that drew the samples.
        :type nn_state: qucumber.nn_states.WaveFunctionBase
        :param samples: A batch of sample states to calculate the observable on.
        :type samples: torch.Tensor
        :returns: The value of the observable of each given basis state.
        :rtype: torch.Tensor
        """
        raise NotImplementedError

    def sample(self, nn_state, k, num_samples=1, initial_state=None, overwrite=False):
        """Draws samples of the *observable* using the given WaveFunction.

        :param nn_state: The WaveFunction to draw samples from.
        :type nn_state: qucumber.nn_states.WaveFunctionBase
        :param k: The number of Gibbs Steps to perform before drawing a sample.
        :type k: int
        :param num_samples: The number of samples to draw.
        :type num_samples: int
        :param initial_state: The initial state of the Markov Chain. If given,
                              `num_samples` will be ignored.
        :type initial_state: torch.Tensor
        :param overwrite: Whether to overwrite the initial_state tensor, if
                          provided, with the updated state of the Markov chain.
        :type overwrite: bool
        :returns: The samples drawn through this observable.
        :rtype: torch.Tensor
        """
        return self.apply(
            nn_state,
            nn_state.sample(
                k=k,
                num_samples=num_samples,
                initial_state=initial_state,
                overwrite=overwrite,
            ),
        )

    def statistics(self, nn_state, num_samples, num_chains=0, burn_in=1000, steps=1):
        """Estimates the expected value, variance, and the standard error of the
        observable over the distribution defined by the WaveFunction.

        :param nn_state: The WaveFunction to draw samples from.
        :type nn_state: qucumber.nn_states.WaveFunctionBase
        :param num_samples: The number of samples to draw. The actual number of
                            samples drawn may be slightly higher if
                            `num_samples % num_chains != 0`.
        :type num_samples: int
        :param num_chains: The number of Markov chains to run in parallel;
                           if 0 or greater than `num_samples`, will use a
                           number of chains equal to `num_samples`. This is not
                           recommended in the case where a `num_samples` is
                           large, as this may use up all the available memory.
        :type num_chains: int
        :param burn_in: The number of Gibbs Steps to perform before recording
                        any samples.
        :type burn_in: int
        :param steps: The number of Gibbs Steps to take between each sample.
        :type steps: int
        :returns: A dictionary containing the (estimated) expected value
                  (key: "mean"), variance (key: "variance"), and standard error
                  (key: "std_error") of the observable.
        :rtype: dict(str, float)
        """
        running_mean = 0.0
        running_variance = 0.0
        running_length = 0

        chains = None
        num_chains = min(num_chains, num_samples) if num_chains != 0 else num_samples
        num_time_steps = int(np.ceil(num_samples / num_chains))
        for i in range(num_time_steps):
            num_gibbs_steps = burn_in if i == 0 else steps

            chains = nn_state.sample(
                num_samples=num_chains,
                k=num_gibbs_steps,
                initial_state=chains,
                overwrite=True,
            )

            sample_stats = self.statistics_from_samples(nn_state, chains)

            running_mean, running_variance, running_length = _update_statistics(
                running_mean,
                running_variance,
                running_length,
                sample_stats["mean"],
                sample_stats["variance"],
                num_chains,
            )

        std_error = np.sqrt(running_variance / running_length)

        return {
            "mean": running_mean,
            "variance": running_variance,
            "std_error": std_error,
        }

    def statistics_from_samples(self, nn_state, samples):
        """Estimates the expected value, variance, and the standard error of the
        observable using the given samples.

        :param nn_state: The WaveFunction that drew the samples.
        :type nn_state: qucumber.nn_states.WaveFunctionBase
        :param samples: A batch of sample states to calculate the observable on.
        :type samples: torch.Tensor
        :returns: A dictionary containing the (estimated) expected value
                  (key: "mean"), variance (key: "variance"), and standard error
                  (key: "std_error") of the observable.
        :rtype: dict(str, float)
        """
        obs_samples = self.apply(nn_state, samples).data

        mean = obs_samples.mean().item()
        variance = obs_samples.var().item()
        std_error = np.sqrt(variance / len(obs_samples))

        return {"mean": mean, "variance": variance, "std_error": std_error}


# make module path show up properly in sphinx docs
ObservableBase.__module__ = "qucumber.observables"


class SumObservable(ObservableBase):
    def __init__(self, o1, o2, name=None, symbol=None):
        if not isinstance(o1, (float, int, ObservableBase)):
            raise TypeError("o1 does not have the right type!")
        if not isinstance(o2, (float, int, ObservableBase)):
            raise TypeError("o2 does not have the right type!")

        self.left = o1
        self.right = o2

        if symbol is None:
            self.symbol = "(" + str(self.left) + " + " + str(self.right) + ")"
        else:
            self.symbol = symbol
        if name is None:
            self.name = "(" + repr(self.left) + " + " + repr(self.right) + ")"
        else:
            self.name = name

    def apply(self, samples, rbm):
        result = 0.0
        if isinstance(self.left, (float, int)):
            result += self.left
        if isinstance(self.right, (float, int)):
            result += self.right

        if isinstance(self.left, ObservableBase):
            result = result + self.left.apply(samples, rbm)
        if isinstance(self.right, ObservableBase):
            result = result + self.right.apply(samples, rbm)

        return result


class ProdObservable(ObservableBase):
    def __init__(self, o1, o2, name=None, symbol=None):
        if not isinstance(o1, (float, int, ObservableBase)):
            raise TypeError("o1 does not have the right type!")
        if not isinstance(o2, (float, int, ObservableBase)):
            raise TypeError("o2 does not have the right type!")

        # assign scalar value to self.left and the observable to self.right
        if isinstance(o1, (float, int)) and isinstance(o2, ObservableBase):
            self.left = o1
            self.right = o2
        elif isinstance(o2, (float, int)) and isinstance(o1, ObservableBase):
            self.left = o2
            self.right = o1
        else:
            raise ValueError("Exactly one of o1 or o2 must be an Observable!")

        if symbol is None:
            self.symbol = "(" + str(self.left) + " * " + str(self.right) + ")"
        else:
            self.symbol = symbol
        if name is None:
            self.name = "(" + repr(self.left) + " * " + repr(self.right) + ")"
        else:
            self.name = name

    def apply(self, samples, rbm):
        return self.left * self.right.apply(samples, rbm)



# Copyright 2019 PIQuIL - All Rights Reserved.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import numpy as np
import torch
from torch import nn
from torch.nn import functional as F
from torch.nn.utils import parameters_to_vector

from qucumber import _warn_on_missing_gpu


class BinaryRBM(nn.Module):
    def __init__(self, num_visible, num_hidden, zero_weights=False, gpu=True):
        super().__init__()
        self.num_visible = int(num_visible)
        self.num_hidden = int(num_hidden)
        self.num_pars = (
            (self.num_visible * self.num_hidden) + self.num_visible + self.num_hidden
        )

        _warn_on_missing_gpu(gpu)
        self.gpu = gpu and torch.cuda.is_available()

        self.device = torch.device("cuda") if self.gpu else torch.device("cpu")

        self.initialize_parameters(zero_weights=zero_weights)

    def __repr__(self):
        return "BinaryRBM(num_visible={}, num_hidden={}, gpu={})".format(
            self.num_visible, self.num_hidden, self.gpu
        )

    def initialize_parameters(self, zero_weights=False):
        """Randomize the parameters of the RBM"""

        gen_tensor = torch.zeros if zero_weights else torch.randn
        self.weights = nn.Parameter(
            (
                gen_tensor(
                    self.num_hidden,
                    self.num_visible,
                    device=self.device,
                    dtype=torch.double,
                )
                / np.sqrt(self.num_visible)
            ),
            requires_grad=False,
        )

        self.visible_bias = nn.Parameter(
            torch.zeros(self.num_visible, device=self.device, dtype=torch.double),
            requires_grad=False,
        )
        self.hidden_bias = nn.Parameter(
            torch.zeros(self.num_hidden, device=self.device, dtype=torch.double),
            requires_grad=False,
        )

    def effective_energy(self, v):
        r"""The effective energies of the given visible states.

        .. math::

            \mathcal{E}(\bm{v}) &= -\sum_{j}b_j v_j
                        - \sum_{i}\log
                            \left\lbrack 1 +
                                  \exp\left(c_{i} + \sum_{j} W_{ij} v_j\right)
                            \right\rbrack

        :param v: The visible states.
        :type v: torch.Tensor

        :returns: The effective energies of the given visible states.
        :rtype: torch.Tensor
        """
        v = v.to(self.weights)
        if len(v.shape) < 2:
            v = v.unsqueeze(0)

        visible_bias_term = torch.mv(v, self.visible_bias)
        hid_bias_term = F.softplus(F.linear(v, self.weights, self.hidden_bias)).sum(1)

        return -(visible_bias_term + hid_bias_term)

    def effective_energy_gradient(self, v, reduce=True):
        """The gradients of the effective energies for the given visible states.

        :param v: The visible states.
        :type v: torch.Tensor
        :param reduce: If `True`, will sum over the gradients resulting from
                       each visible state. Otherwise will return a batch of
                       gradient vectors.

        :returns: Will return a vector (or matrix if `reduce=False` and multiple
                  visible states were given as a matrix) containing the gradients
                  for all parameters (computed on the given visible states v).
        :rtype: torch.Tensor
        """
        v = v.to(self.weights)
        prob = self.prob_h_given_v(v)

        if v.dim() < 2:
            W_grad = -torch.ger(prob, v)
            vb_grad = -v
            hb_grad = -prob
        else:
            if reduce:
                W_grad = -torch.matmul(prob.t(), v)
                vb_grad = -torch.sum(v, 0)
                hb_grad = -torch.sum(prob, 0)
            else:
                W_grad = -torch.einsum("ij,ik->ijk", prob, v)
                vb_grad = -v
                hb_grad = -prob
                vec = [W_grad.view(v.size()[0], -1), vb_grad, hb_grad]
                return torch.cat(vec, dim=1)

        return parameters_to_vector([W_grad, vb_grad, hb_grad])

    def prob_v_given_h(self, h, out=None):
        """Given a hidden unit configuration, compute the probability
        vector of the visible units being on.

        :param h: The hidden unit
        :type h: torch.Tensor
        :param out: The output tensor to write to.
        :type out: torch.Tensor

        :returns: The probability of visible units being active given the
                  hidden state.
        :rtype: torch.Tensor
        """
        if h.dim() < 2:  # create extra axis, if needed
            h = h.unsqueeze(0)
            unsqueezed = True
        else:
            unsqueezed = False

        p = torch.addmm(
            self.visible_bias.data, h, self.weights.data, out=out
        ).sigmoid_()

        if unsqueezed:
            return p.squeeze_(0)  # remove superfluous axis, if it exists
        else:
            return p

    def prob_h_given_v(self, v, out=None):
        """Given a visible unit configuration, compute the probability
        vector of the hidden units being on.

        :param h: The hidden unit.
        :type h: torch.Tensor
        :param out: The output tensor to write to.
        :type out: torch.Tensor

        :returns: The probability of hidden units being active given the
                  visible state.
        :rtype: torch.Tensor
        """
        if v.dim() < 2:  # create extra axis, if needed
            v = v.unsqueeze(0)
            unsqueezed = True
        else:
            unsqueezed = False

        p = torch.addmm(
            self.hidden_bias.data, v, self.weights.data.t(), out=out
        ).sigmoid_()

        if unsqueezed:
            return p.squeeze_(0)  # remove superfluous axis, if it exists
        else:
            return p

    def sample_v_given_h(self, h, out=None):
        """Sample/generate a visible state given a hidden state.

        :param h: The hidden state.
        :type h: torch.Tensor
        :param out: The output tensor to write to.
        :type out: torch.Tensor

        :returns: The sampled visible state.
        :rtype: torch.Tensor
        """
        v = self.prob_v_given_h(h, out=out)
        v = torch.bernoulli(v, out=out)  # overwrite v with its sample
        return v

    def sample_h_given_v(self, v, out=None):
        """Sample/generate a hidden state given a visible state.

        :param h: The visible state.
        :type h: torch.Tensor
        :param out: The output tensor to write to.
        :type out: torch.Tensor

        :returns: The sampled hidden state.
        :rtype: torch.Tensor
        """
        h = self.prob_h_given_v(v, out=out)
        h = torch.bernoulli(h, out=out)  # overwrite h with its sample
        return h

    def gibbs_steps(self, k, initial_state, overwrite=False):
        r"""Performs k steps of Block Gibbs sampling. One step consists of sampling
        the hidden state :math:`\bm{h}` from the conditional distribution
        :math:`p(\bm{h}\:|\:\bm{v})`, and sampling the visible
        state :math:`\bm{v}` from the conditional distribution
        :math:`p(\bm{v}\:|\:\bm{h})`.

        :param k: Number of Block Gibbs steps.
        :type k: int
        :param initial_state: The initial state of the Markov Chains.
        :type initial_state: torch.Tensor
        :param overwrite: Whether to overwrite the initial_state tensor, if it is provided.
        :type overwrite: bool
        """
        v = initial_state.to(device=self.device, dtype=torch.double)

        if overwrite is False:
            v = v.clone()

        h = torch.zeros(
            v.shape[0], self.num_hidden, device=self.device, dtype=torch.double
        )

        for _ in range(k):
            self.sample_h_given_v(v, out=h)
            self.sample_v_given_h(h, out=v)

        return v

    def partition(self, space):
        """Compute the partition function of the RBM.

        :param space: A rank 2 tensor of the visible space.
        :type space: torch.Tensor

        :returns: The value of the partition function evaluated at the current
                  state of the RBM.
        :rtype: torch.Tensor
        """
        neg_free_energies = -self.effective_energy(space)
        logZ = neg_free_energies.logsumexp(0)
        return logZ.exp()



# Copyright 2019 PIQuIL - All Rights Reserved.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import time

from .callback import CallbackBase


class Timer(CallbackBase):
    """Callback which records the training time.

    This callback is always called at the start and end of training. It will
    run at the end of an epoch or batch if the given model's `stop_training`
    property is set to True.

    :param verbose: Whether to print the elapsed time at the end of training.
    :type verbose: bool
    """

    def __init__(self, verbose=True):
        self.verbose = verbose
        self.already_notified = False

    def on_train_start(self, nn_state):
        self.start_time = time.time()

    def on_batch_end(self, nn_state, epoch, batch):
        if nn_state.stop_training:
            if self.verbose and not self.already_notified:
                print(f"Training terminated at epoch: {epoch}, batch: {batch}")
                self.already_notified = True

    def on_epoch_end(self, nn_state, epoch):
        if nn_state.stop_training:
            if self.verbose and not self.already_notified:
                print(f"Training terminated at epoch: {epoch}")
                self.already_notified = True

    def on_train_end(self, nn_state):
        self.calculate_elapsed_time()

    def calculate_elapsed_time(self):
        self.end_time = time.time()
        self.training_time = self.end_time - self.start_time
        if self.verbose:
            print(
                "Total time elapsed during training: {:6.3f} s".format(
                    self.training_time
                )
            )



# Copyright 2019 PIQuIL - All Rights Reserved.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import torch


def make_complex(x, y=None):
    """A function that combines the real (x) and imaginary (y) parts of a
    vector or a matrix.

    .. note:: x and y must have the same shape. Also, this will not work for
              rank zero tensors.

    :param x: The real part
    :type x: torch.Tensor
    :param y: The imaginary part. Can be None, in which case, the resulting
              complex tensor will have imaginary part equal to zero.
    :type y: torch.Tensor

    :returns: The tensor [x,y] = x + yi.
    :rtype: torch.Tensor
    """
    if y is None:
        y = torch.zeros_like(x)
    return torch.cat((x.unsqueeze(0), y.unsqueeze(0)), dim=0)


def real(x):
    """Returns the real part of a complex tensor.

    :param x: The complex tensor
    :type x: torch.Tensor

    :returns: The real part of `x`; will have one less dimension than `x`.
    :rtype: torch.Tensor
    """
    return x[0, ...]


def imag(x):
    """Returns the imaginary part of a complex tensor.

    :param x: The complex tensor
    :type x: torch.Tensor

    :returns: The imaginary part of `x`; will have one less dimension than `x`.
    :rtype: torch.Tensor
    """
    return x[1, ...]


def scalar_mult(x, y, out=None):
    """A function that computes the product between complex matrices and scalars,
    complex vectors and scalars or two complex scalars.

    :param x: A complex scalar, vector or matrix.
    :type x: torch.Tensor
    :param y: A complex scalar, vector or matrix.
    :type y: torch.Tensor

    :returns: The product between `x` and `y`.
              Either overwrites `out`, or returns a new tensor.
    :rtype: torch.Tensor
    """
    if out is None:
        out = torch.zeros(
            2, *((x[0] * y[0]).size()), dtype=torch.double, device=x.device
        )
    else:
        if out is x or out is y:
            raise RuntimeError("Can't overwrite an argument!")

    out[0] = (x[0] * y[0]) - (x[1] * y[1])
    out[1] = (x[0] * y[1]) + (x[1] * y[0])

    return out


def matmul(x, y):
    """A function that computes complex matrix-matrix and matrix-vector products.

    .. note:: If one wishes to do matrix-vector products, the vector must be
              the second argument (y).

    :param x: A complex matrix.
    :type x: torch.Tensor
    :param y: A complex vector or matrix.
    :type y: torch.Tensor

    :returns: The product between x and y.
    :rtype: torch.Tensor
    """
    if len(list(y.size())) == 2:
        # if one of them is a vector (i.e. wanting to do MV mult)
        z = torch.zeros(2, x.size()[1], dtype=torch.double, device=x.device)
        z[0] = torch.mv(x[0], y[0]) - torch.mv(x[1], y[1])
        z[1] = torch.mv(x[0], y[1]) + torch.mv(x[1], y[0])

    if len(list(y.size())) == 3:
        z = torch.zeros(
            2, x.size()[1], y.size()[2], dtype=torch.double, device=x.device
        )
        z[0] = torch.matmul(x[0], y[0]) - torch.matmul(x[1], y[1])
        z[1] = torch.matmul(x[0], y[1]) + torch.matmul(x[1], y[0])

    return z


def inner_prod(x, y):
    """A function that returns the inner product of two complex vectors,
    x and y (<x|y>).

    :param x: A complex vector.
    :type x: torch.Tensor
    :param y: A complex vector.
    :type y: torch.Tensor

    :raises ValueError: If x and y are not complex vectors with their first
                        dimensions being 2, then the function will not execute.

    :returns: The inner product, :math:`\\langle x\\vert y\\rangle`.
    :rtype: torch.Tensor
    """
    z = torch.zeros(2, dtype=torch.double, device=x.device)

    if len(list(x.size())) == 2 and len(list(y.size())) == 2:
        z[0] = torch.dot(x[0], y[0]) + torch.dot(x[1], y[1])
        z[1] = torch.dot(x[0], y[1]) + torch.dot(-x[1], y[0])

    if len(list(x.size())) == 1 and len(list(y.size())) == 1:
        z[0] = (x[0] * y[0]) + (x[1] * y[1])
        z[1] = (x[0] * y[1]) + (-x[1] * y[0])

    return z


def outer_prod(x, y):
    """A function that returns the outer product of two complex vectors, x
    and y.

    :param x: A complex vector.
    :type x: torch.Tensor
    :param y: A complex vector.
    :type y: torch.Tensor

    :raises ValueError:	If x and y are not complex vectors with their first
                        dimensions being 2, then an error will be raised.

    :returns: The outer product between x and y,
        :math:`\\vert x \\rangle\\langle y\\vert`.
    :rtype: torch.Tensor
    """
    if len(list(x.size())) != 2 or len(list(y.size())) != 2:
        raise ValueError("An input is not of the right dimension.")

    z = torch.zeros(2, x.size()[1], y.size()[1], dtype=torch.double, device=x.device)
    z[0] = torch.ger(x[0], y[0]) - torch.ger(x[1], -y[1])
    z[1] = torch.ger(x[0], -y[1]) + torch.ger(x[1], y[0])

    return z


def conjugate(x):
    """A function that takes the conjugate transpose of the argument.

    :param x: A complex vector or matrix.
    :type x: torch.Tensor

    :returns: The conjugate of x.
    :rtype: torch.Tensor
    """
    if x.dim() == 1 or x.dim() == 2:
        z = torch.zeros_like(x, dtype=torch.double, device=x.device)
        z[0] = x[0]
        z[1] = -x[1]

    if x.dim() == 3:
        z = torch.zeros(
            2, x.size()[2], x.size()[1], dtype=torch.double, device=x.device
        )
        z[0] = torch.transpose(x[0], 0, 1)
        z[1] = -torch.transpose(x[1], 0, 1)

    return z


def elementwise_mult(x, y):
    """Alias for :func:`scalar_mult`."""
    return scalar_mult(x, y)


def elementwise_division(x, y):
    """Elementwise division of x by y.

    :param x: A complex tensor.
    :type x: torch.Tensor
    :param y: A complex tensor.
    :type y: torch.Tensor

    :rtype: torch.Tensor
    """
    if x.shape != y.shape:
        raise ValueError("x and y must have the same shape!")

    y_star = y.clone()
    y_star[1] *= -1

    sqrd_abs_y = absolute_value(y).pow_(2)

    return elementwise_mult(x, y_star).div_(sqrd_abs_y)


def absolute_value(x):
    """Computes the complex absolute value elementwise.

    :param x: A complex tensor.
    :type x: torch.Tensor

    :returns: A real tensor.
    :rtype: torch.Tensor
    """
    x_star = x.clone()
    x_star[1] *= -1
    return elementwise_mult(x, x_star)[0].sqrt_()


def kronecker_prod(x, y):
    """A function that returns the tensor / kronecker product of 2 complex
    tensors, x and y.

    :param x: A complex matrix.
    :type x: torch.Tensor
    :param y: A complex matrix.
    :type y: torch.Tensor

    :raises ValueError: If x and y do not have 3 dimensions or their first
                        dimension is not 2, the function cannot execute.

    :returns: The tensorproduct of x and y, :math:`x \\otimes y`.
    :rtype: torch.Tensor
    """
    if len(list(x.size())) != 3 or len(list(y.size())) != 3:
        raise ValueError("An input is not of the right dimension.")

    z = torch.zeros(
        2,
        x.size()[1] * y.size()[1],
        x.size()[2] * y.size()[2],
        dtype=torch.double,
        device=x.device,
    )

    row_count = 0

    for i in range(x.size()[1]):
        for k in range(y.size()[1]):
            column_count = 0
            for j in range(x.size()[2]):
                for l in range(y.size()[2]):

                    z[0][row_count][column_count] = (x[0][i][j] * y[0][k][l]) - (
                        x[1][i][j] * y[1][k][l]
                    )
                    z[1][row_count][column_count] = (x[0][i][j] * y[1][k][l]) + (
                        x[1][i][j] * y[0][k][l]
                    )

                    column_count += 1
            row_count += 1

    return z


def scalar_divide(x, y):
    """A function that computes the division of x by y.

    :param x: The numerator (a complex scalar, vector or matrix).
    :type x: torch.Tensor
    :param y: The denominator (a complex scalar).
    :type y: torch.Tensor

    :returns: x / y
    :rtype: torch.Tensor
    """
    y_star = conjugate(y)
    numerator = scalar_mult(x, y_star)
    denominator = real(scalar_mult(y, y_star))

    return numerator / denominator


def norm_sqr(x):
    """A function that returns the squared norm of the argument.

    :param x: A complex scalar.
    :type x: torch.Tensor

    :returns: :math:`|x|^2`.
    :rtype: torch.Tensor
    """
    return real(inner_prod(x, x))


def norm(x):
    """A function that returns the norm of the argument.

    :param x: A complex scalar.
    :type x: torch.Tensor

    :returns: :math:`|x|`.
    :rtype: torch.Tensor
    """
    return norm_sqr(x).sqrt_()



# Copyright 2019 PIQuIL - All Rights Reserved.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from collections.abc import MutableSequence

from .callback import CallbackBase


class CallbackList(CallbackBase, MutableSequence):
    def __init__(self, callbacks):
        super().__init__()
        self.callbacks = list(callbacks)

    def __len__(self):
        return len(self.callbacks)

    def __getitem__(self, key):
        return self.callbacks[key]

    def __setitem__(self, key, value):
        if isinstance(value, CallbackBase):
            self.callbacks[key] = value
        else:
            raise TypeError(
                "value must be an instance of qucumber.callbacks.CallbackBase"
            )

    def __delitem__(self, index):
        del self.callbacks[index]

    def __iter__(self):
        return iter(self.callbacks)

    def __add__(self, other):
        return CallbackList(self.callbacks + other.callbacks)

    def insert(self, index, value):
        if isinstance(value, CallbackBase):
            self.callbacks.insert(index, value)
        else:
            raise TypeError(
                "value must be an instance of qucumber.callbacks.CallbackBase"
            )

    def on_train_start(self, rbm):
        for cb in self.callbacks:
            cb.on_train_start(rbm)

    def on_train_end(self, rbm):
        for cb in self.callbacks:
            cb.on_train_end(rbm)

    def on_epoch_start(self, rbm, epoch):
        for cb in self.callbacks:
            cb.on_epoch_start(rbm, epoch)

    def on_epoch_end(self, rbm, epoch):
        for cb in self.callbacks:
            cb.on_epoch_end(rbm, epoch)

    def on_batch_start(self, rbm, epoch, batch):
        for cb in self.callbacks:
            cb.on_batch_start(rbm, epoch, batch)

    def on_batch_end(self, rbm, epoch, batch):
        for cb in self.callbacks:
            cb.on_batch_end(rbm, epoch, batch)



# Copyright 2019 PIQuIL - All Rights Reserved.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import numpy as np


from .callback import CallbackBase
from .metric_evaluator import MetricEvaluator
from .observable_evaluator import ObservableEvaluator


class EarlyStopping(CallbackBase):
    r"""Stop training once the model stops improving.

    There are three different stopping criteria available:

    `relative`, which computes the relative change between the two model
    evaluation steps:

    .. math:: \left\vert\frac{M_{t-p} - M_t}{M_{t-p}}\right\vert < \epsilon

    `absolute` computes the absolute change:

    .. math:: \left\vert M_{t-p} - M_t\right\vert < \epsilon

    `variance` computes the absolute change, but scales the change by the
    standard deviation of the quantity of interest, such that the tolerance,
    `\epsilon` can now be interpreted as the "number of standard deviations":

    .. math:: \left\vert\frac{M_{t-p} - M_t}{\sigma_{t-p}}\right\vert < \epsilon

    where :math:`M_t` is the metric value at the current evaluation
    (time :math:`t`), :math:`p` is the "patience" parameter,
    :math:`\sigma_t` is the standard deviation of the metric, and
    :math:`\epsilon` is the tolerance.

    This callback is called at the end of each epoch.

    :param period: Frequency with which the callback checks whether training
                   has converged (in epochs).
    :type period: int
    :param tolerance: The maximum relative change required to consider training
                      as having converged.
    :type tolerance: float
    :param patience: How many intervals to wait before claiming the training
                     has converged.
    :type patience: int
    :param evaluator_callback: An instance of
        :class:`MetricEvaluator<MetricEvaluator>` or
        :class:`ObservableEvaluator<ObservableEvaluator>` which computes the metric
        that we want to check for convergence.
    :type evaluator_callback: :class:`MetricEvaluator<MetricEvaluator>` or
                              :class:`ObservableEvaluator<ObservableEvaluator>`
    :param quantity_name: The name of the metric/observable stored in `evaluator_callback`.
    :type quantity_name: str
    :param criterion: The stopping criterion to use. Must be one of the following:
                      `relative`, `absolute`, `variance`.
    :type criterion: str
    """

    def __init__(
        self,
        period,
        tolerance,
        patience,
        evaluator_callback,
        quantity_name,
        criterion="relative",
    ):
        self.period = period
        self.tolerance = tolerance
        self.patience = int(patience)
        self.quantity_name = quantity_name

        if isinstance(evaluator_callback, MetricEvaluator):
            self.evaluator_callback = evaluator_callback
            self.value_getter = self.evaluator_callback.get_value

            if criterion == self._convergence_criteria[2]:
                raise TypeError(
                    "Can't use a variance based convergence criterion "
                    "with MetricEvaluator!"
                )
        elif isinstance(evaluator_callback, ObservableEvaluator):
            self.evaluator_callback = evaluator_callback
            self.value_getter = lambda *args: self.evaluator_callback.get_value(*args)[
                "mean"
            ]
            self.variance_getter = lambda *args: self.evaluator_callback.get_value(
                *args
            )["variance"]
        else:
            raise TypeError(
                "evaluator_callback must be an instance of "
                "either MetricEvaluator or ObservableEvaluator!"
            )

        convergence_criteria = {
            "relative": self._relative_change,
            "absolute": self._absolute_change,
            "variance": self._variance_scaled_abs_change,
        }

        try:
            self.criterion = criterion.strip().lower()
        except KeyError:
            raise ValueError(
                "criterion must be one of " + ", ".join(convergence_criteria.keys())
            )

        self.deviation = convergence_criteria[self.criterion]
        self.last_epoch = None

    def _change_in_metric(self):
        return self.value_getter(
            self.quantity_name, -self.patience
        ) - self.value_getter(self.quantity_name)

    def _relative_change(self):
        relative_change = self._change_in_metric() / self.value_getter(
            self.quantity_name, -self.patience
        )
        return abs(relative_change)

    def _absolute_change(self):
        return abs(self._change_in_metric())

    def _variance_scaled_abs_change(self):
        return abs(self._change_in_metric()) / np.sqrt(
            self.variance_getter(self.quantity_name, -self.patience)
        )

    def on_epoch_end(self, nn_state, epoch):
        if epoch % self.period == 0:
            if len(self.evaluator_callback) >= self.patience:
                if self.deviation() < self.tolerance:
                    nn_state.stop_training = True
                    self.last_epoch = epoch



# Copyright 2019 PIQuIL - All Rights Reserved.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from operator import itemgetter

import numpy as np

import matplotlib.pyplot as plt
from matplotlib import ticker

from .callback import CallbackBase


class LivePlotting(CallbackBase):
    """Plots metrics/observables.

    This callback is called at the end of each epoch.

    :param period: Frequency with which the callback updates the plots
                   (in epochs).
    :type period: int
    :param evaluator_callback: An instance of
        :class:`MetricEvaluator<MetricEvaluator>` or
        :class:`ObservableEvaluator<ObservableEvaluator>`
        which computes the metric/observable that we want to plot.
    :type evaluator_callback: :class:`MetricEvaluator<MetricEvaluator>` or
                           :class:`ObservableEvaluator<ObservableEvaluator>`
    :param quantity_name: The name of the metric/observable stored in `evaluator_callback`.
    :type quantity_name: str
    :param error_name: The name of the error stored in `evaluator_callback`.
    :type error_name: str
    """

    def __init__(
        self,
        period,
        evaluator_callback,
        quantity_name,
        error_name=None,
        total_epochs=None,
        smooth=True,
    ):
        self.period = period
        self.evaluator_callback = evaluator_callback
        self.quantity_name = quantity_name
        self.error_name = error_name
        self.last_epoch = 0
        self.total_epochs = total_epochs
        self.smooth = smooth

    def on_train_start(self, nn_state):
        self.fig, self.ax = plt.subplots()

        if self.total_epochs:
            self.ax.set_xlim(0, self.total_epochs)

        self.ax.grid()
        self.ax.xaxis.set_major_locator(ticker.MultipleLocator(min(self.period, 5.0)))
        self.fig.show()
        self.fig.canvas.draw()

    def on_epoch_end(self, nn_state, epoch):
        if epoch % self.period == 0:
            self.last_epoch = epoch

            epochs = np.array(
                list(map(itemgetter(0), self.evaluator_callback.past_values))
            )

            past_values = np.array(
                list(
                    map(
                        itemgetter(self.quantity_name),
                        map(itemgetter(1), self.evaluator_callback.past_values),
                    )
                )
            )

            self.ax.clear()
            p = self.ax.plot(epochs, past_values)

            if self.error_name is not None:
                std_error = np.array(
                    list(
                        map(
                            itemgetter(self.error_name),
                            map(itemgetter(1), self.evaluator_callback.past_values),
                        )
                    )
                )

                lower = past_values - std_error
                upper = past_values + std_error

                self.ax.fill_between(
                    epochs, lower, upper, color=p[0].get_color(), alpha=0.4
                )

            y_avg = np.max(np.abs(past_values))
            y_log_avg = np.log10(y_avg) if y_avg != 0 else -1.0
            y_tick_exp = int(np.sign(y_log_avg) * np.ceil(np.abs(y_log_avg)))
            y_tick_interval = (10 ** y_tick_exp) / 2.0
            self.ax.yaxis.set_major_locator(ticker.MultipleLocator(y_tick_interval))

            self.ax.set_xlabel("Epochs")
            self.ax.set_ylabel(self.quantity_name)
            self.ax.grid()
            self.fig.canvas.draw()

    def on_train_end(self, nn_state):
        self.on_epoch_end(nn_state, self.last_epoch)



# Copyright 2019 PIQuIL - All Rights Reserved.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import csv

import numpy as np

from .callback import CallbackBase
from qucumber.observables import System


class ObservableStatistics:
    """A data structure which allows easy access to past values of Observable statistics.

    :param data: The historical statistics of an Observable.
    :type data: list[dict(str, float)]
    """

    def __init__(self, data):
        self.data = data

    def __getattr__(self, statistic):
        """Return an array containing values of the given statistic.

        :param statistic: The statistic to retrieve.
        :type statistic: str

        :returns: The past values of the statistic.
        :rtype: np.array
        """
        try:
            stat = statistic[:-1] if statistic.endswith("s") else statistic

            if len(self.data) > 0 and stat in self.data[0].keys():
                return np.array([stat_dict[stat] for stat_dict in self.data])

            return np.array([stat_dict[statistic] for stat_dict in self.data])
        except KeyError:
            raise AttributeError(
                "'{}' is not a statistic being tracked by this object.".format(
                    statistic
                )
            )

    def __getitem__(self, statistic):
        """Alias for
        :func:`__getattr__<qucumber.callbacks.observable_evaluator.ObservableStatistics.__getattr__>`
        to enable subscripting."""
        return self.__getattr__(statistic)


class ObservableEvaluator(CallbackBase):
    r"""Evaluate and hold on to the results of the given observable(s).

    This callback is called at the end of each epoch.

    .. note::
        Since callback are given to :func:`fit<qucumber.nn_states.WaveFunctionBase.fit>`
        as a list, they will be called in a deterministic order. It is
        therefore recommended that instances of
        :class:`ObservableEvaluator<ObservableEvaluator>` be among the first callbacks in
        the list passed to :func:`fit<qucumber.nn_states.WaveFunctionBase.fit>`,
        as one would often use it in conjunction with other callbacks like
        :class:`EarlyStopping<EarlyStopping>` which may depend on
        :class:`ObservableEvaluator<ObservableEvaluator>` having been called.

    :param period: Frequency with which the callback evaluates the given
                   observables(s).
    :type period: int
    :param observables: A list of Observables. Observable statistics are
                        evaluated by sampling the WaveFunction. Note that
                        observables that have the same name will conflict,
                        and precedence will be given to the one which appears
                        later in the list.
    :type observables: list(qucumber.observables.ObservableBase)
    :param verbose: Whether to print metrics to stdout.
    :type verbose: bool
    :param log: A filepath to log metric values to in CSV format.
    :type log: str
    :param \**sampling_kwargs: Keyword arguments to be passed to `Observable.statistics`.
                               Ex. `num_samples`, `num_chains`, `burn_in`, `steps`.
    """

    def __init__(self, period, observables, verbose=False, log=None, **sampling_kwargs):
        self.period = period
        self.past_values = []
        self.system = System(*observables)
        self.sampling_kwargs = sampling_kwargs
        self.last = {}
        self.verbose = verbose
        self.log = log

        self.csv_fields = ["epoch"]
        for obs_name in self.system.observables.keys():
            self.csv_fields.append(obs_name + "_mean")
            self.csv_fields.append(obs_name + "_variance")
            self.csv_fields.append(obs_name + "_std_error")

        if self.log is not None:
            with open(self.log, "a") as log_file:
                writer = csv.DictWriter(log_file, fieldnames=self.csv_fields)
                writer.writeheader()

    def __len__(self):
        """Return the number of timesteps that observables have been evaluated for.

        :rtype: int
        """
        return len(self.past_values)

    def __getattr__(self, observable):
        """Return an ObservableStatistics containing recorded statistics of the given observable.

        :param observable: The observable to retrieve.
        :type observable: str

        :returns: The past values of the observable.
        :rtype: ObservableStatistics
        """
        try:
            return ObservableStatistics(
                [values[observable] for _, values in self.past_values]
            )
        except KeyError:
            raise AttributeError(
                "'{}' is not an Observable being tracked by this object.".format(
                    observable
                )
            )

    def __getitem__(self, observable):
        """Alias for :func:`__getattr__<qucumber.callbacks.ObservableEvaluator.__getattr__>`
        to enable subscripting."""
        return self.__getattr__(observable)

    @property
    def epochs(self):
        """Return a list of all epochs that have been recorded.

        :rtype: np.array
        """
        return np.array([epoch for epoch, _ in self.past_values])

    @property
    def names(self):
        """The names of the tracked observables.

        :rtype: list[str]
        """
        return list(self.system.observables.keys())

    def clear_history(self):
        """Delete all statistics the instance is currently storing."""
        self.past_values = []
        self.last = {}

    def get_value(self, name, index=None):
        """Retrieve the statistics of the desired observable from the given timestep.

        :param name: The name of the observable to retrieve.
        :type name: str
        :param index: The index/timestep from which to retrieve the observable.
                      Negative indices are supported. If None, will just get
                      the most recent value.
        :type index: int or None

        :rtype: dict(str, float)
        """
        index = index if index is not None else -1
        return self.past_values[index][-1][name]

    def on_epoch_end(self, nn_state, epoch):
        if epoch % self.period == 0:
            obs_vals = self.system.statistics(nn_state, **self.sampling_kwargs)

            self.last = obs_vals.copy()
            self.past_values.append((epoch, obs_vals))

            if self.verbose is True:
                print(f"Epoch: {epoch}\n", end="", flush=True)
                partially_formatted = {
                    k: "\t".join(f"{s}: {sv:.6f}" for s, sv in stats.items())
                    for k, stats in self.last.items()
                }
                print(
                    "\n".join(
                        f"  {k}:\n    {stats}"
                        for k, stats in partially_formatted.items()
                    )
                )

            if self.log is not None:
                row = {"epoch": epoch}
                for obs_name, obs_stats in self.last.items():
                    for stat_name, stat in obs_stats.items():
                        row[obs_name + "_" + stat_name] = stat

                with open(self.log, "a") as log_file:
                    writer = csv.DictWriter(
                        log_file, fieldnames=self.csv_fields, extrasaction="ignore"
                    )
                    writer.writerow(row)



# Copyright 2019 PIQuIL - All Rights Reserved.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import warnings

import torch

from .__version__ import __version__


def _warn_on_missing_gpu(gpu):
    if gpu and not torch.cuda.is_available():
        warnings.warn("Could not find GPU: will continue with CPU.", ResourceWarning)


def set_random_seed(seed, cpu=True, gpu=False, quiet=False):
    if gpu and torch.cuda.is_available():
        if not quiet:
            warnings.warn(
                "GPU random seeds are not completely deterministic. "
                "Proceed with caution."
            )
        torch.cuda.manual_seed(seed)

    if cpu:
        torch.manual_seed(seed)



# Copyright 2019 PIQuIL - All Rights Reserved.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from .observable import ObservableBase
from qucumber.utils import cplx


def swap(s1, s2, A):
    r"""Applies the Swap operator onto two sets of samples, `s1` and `s2`.

    The sites that will be swapped are defined by `A`.

    :param s1: A rank 2 tensor of spin samples representing the first replica.
    :type s1: torch.Tensor
    :param s2: A rank 2 tensor of spin samples representing the second replica.
    :type s2: torch.Tensor
    :param A: The sites that will be swapped between the two replicas.
    :type A: int or list or np.array or torch.Tensor
    """
    _s = s1[:, A].clone()
    s1[:, A] = s2[:, A]
    s2[:, A] = _s
    return s1, s2


class SWAP(ObservableBase):
    r"""The :math:`\text{Swap}_A` observable.

    Can be used to compute the 2nd Renyi entropy of the region A through:

    :math:`S_2 = -\ln\langle \text{Swap}_A \rangle`

    Ref: PhysRevLett.104.157201

    :param A: The sites contained in the region A.
    :type A: int or list or np.array or torch.Tensor
    """

    def __init__(self, A):
        self.name = "SWAP"
        self.symbol = "S"
        self.A = A

    def apply(self, nn_state, samples):
        r"""Computes the swap operator which an estimator for the 2nd Renyi
        entropy.

        The swap operator requires access to two identical copies of a
        wavefunction. In practice, this translates to the requirement of
        having two independent sets of samples from the wavefunction replicas.
        For this purpose, the batch of samples stored in the param samples is
        split into two subsets. Although this procedure is designed to break
        the autocorrelation between the samples, it must be used with caution.
        For a fully unbiased estimate of the entanglement entropy, the batch
        of samples needs to be built from two independent initializations of
        the wavefunction each having a different random number generator.

        :param nn_state: The WaveFunction that drew the samples.
        :type nn_state: qucumber.nn_states.WaveFunctionBase
        :param samples: A batch of samples to calculate the observable on.
                        Must be using the :math:`\sigma_i = 0, 1` convention.
        :type samples: torch.Tensor
        """
        samples = samples.to(device=nn_state.device, copy=True)

        # split the batch of samples into two equal batches
        # if their total number is odd, the last sample is ignored
        _ns = samples.shape[0] // 2
        samples1 = samples[:_ns, :]
        samples2 = samples[_ns : _ns * 2, :]

        psi_ket1 = nn_state.psi(samples1)
        psi_ket2 = nn_state.psi(samples2)

        psi_ket = cplx.elementwise_mult(psi_ket1, psi_ket2)
        psi_ket_star = cplx.conjugate(psi_ket)

        samples1_, samples2_ = swap(samples1, samples2, self.A)
        psi_bra1 = nn_state.psi(samples1_)
        psi_bra2 = nn_state.psi(samples2_)

        psi_bra = cplx.elementwise_mult(psi_bra1, psi_bra2)
        psi_bra_star = cplx.conjugate(psi_bra)
        return cplx.real(cplx.elementwise_division(psi_bra_star, psi_ket_star))



# Copyright 2019 PIQuIL - All Rights Reserved.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import warnings

from .early_stopping import EarlyStopping


class VarianceBasedEarlyStopping(EarlyStopping):
    r"""
    .. deprecated:: 1.2
       Use :class:`EarlyStopping<EarlyStopping>` instead.

    Stop training once the model stops improving. This is a variation
    on the :class:`EarlyStopping<EarlyStopping>` class which takes the variance
    of the metric into account.

    The specific criterion for stopping is:

    .. math:: \left\vert\frac{M_{t-p} - M_t}{\sigma_{t-p}}\right\vert < \kappa

    where :math:`M_t` is the metric value at the current evaluation
    (time :math:`t`), :math:`p` is the "patience" parameter,
    :math:`\sigma_t` is the standard deviation of the metric, and
    :math:`\kappa` is the tolerance.

    This callback is called at the end of each epoch.

    :param period: Frequency with which the callback checks whether training
                   has converged (in epochs).
    :type period: int
    :param tolerance: The maximum (standardized) change required to consider
                      training as having converged.
    :type tolerance: float
    :param patience: How many intervals to wait before claiming the training
                     has converged.
    :type patience: int
    :param evaluator_callback: An instance of
        :class:`MetricEvaluator<MetricEvaluator>` or
        :class:`ObservableEvaluator<ObservableEvaluator>` which computes the
        metric/observable that we want to check for convergence.
    :type evaluator_callback: :class:`MetricEvaluator<MetricEvaluator>` or
                           :class:`ObservableEvaluator<ObservableEvaluator>`
    :param quantity_name: The name of the metric/obserable stored in `evaluator_callback`.
    :type quantity_name: str
    :param variance_name: The name of the variance stored in `evaluator_callback`.
                          Ignored, exists for backward compatibility.
    :type variance_name: str
    """

    def __init__(
        self,
        period,
        tolerance,
        patience,
        evaluator_callback,
        quantity_name,
        variance_name=None,
    ):
        warnings.warn(
            "VarianceBasedEarlyStopping has been deprecated, and "
            "relevant functionality moved to EarlyStopping.",
            DeprecationWarning,
            2,
        )
        super().__init__(
            period,
            tolerance,
            patience,
            evaluator_callback,
            quantity_name,
            criterion="variance",
        )



from .observable import ObservableBase
from .entanglement import SWAP
from .pauli import SigmaX, SigmaY, SigmaZ
from .interactions import NeighbourInteraction
from .system import System
from .utils import to_01, to_pm1



# Copyright 2019 PIQuIL - All Rights Reserved.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import torch

from qucumber.utils import cplx
from .observable import ObservableBase
from .utils import to_pm1


def flip_spin(i, samples):
    r"""Flip the i-th spin configuration in `samples`.

    :param i: The i-th spin.
    :type i: int
    :param samples: A batch of samples.
                    Must be using the :math:`\sigma_i = 0, 1` convention.
    :type samples: torch.Tensor
    """
    samples[:, i].sub_(1).abs_()


class SigmaX(ObservableBase):
    r"""The :math:`\sigma_x` observable

    Computes the magnetization in the X direction of a spin chain.

    :param absolute: Specifies whether to estimate the absolute magnetization.
    :type absolute: bool
    """

    def __init__(self, absolute=False):
        self.name = "SigmaX"
        self.symbol = "X"
        self.absolute = absolute

    def apply(self, nn_state, samples):
        r"""Computes the magnetization along X of each sample in the given batch of samples.

        Assumes that the computational basis that the WaveFunction was trained
        on was the Z basis.

        :param nn_state: The WaveFunction that drew the samples.
        :type nn_state: qucumber.nn_states.WaveFunctionBase
        :param samples: A batch of samples to calculate the observable on.
                        Must be using the :math:`\sigma_i = 0, 1` convention.
        :type samples: torch.Tensor
        """
        samples = samples.to(device=nn_state.device)

        # vectors of shape: (2, num_samples,)
        psis = nn_state.psi(samples)
        psi_ratio_sum = torch.zeros_like(psis)

        for i in range(samples.shape[-1]):  # sum over spin sites
            flip_spin(i, samples)  # flip the spin at site i

            # compute ratio of psi_(-i) / psi and add it to the running sum
            psi_ratio = nn_state.psi(samples)
            psi_ratio = cplx.elementwise_division(psi_ratio, psis)
            psi_ratio_sum.add_(psi_ratio)

            flip_spin(i, samples)  # flip it back

        # take real part (imaginary part should be approximately zero)
        # and divide by number of spins
        res = cplx.real(psi_ratio_sum).div_(samples.shape[-1])
        if self.absolute:
            return res.abs_()
        else:
            return res


class SigmaY(ObservableBase):
    r"""The :math:`\sigma_y` observable

    Computes the magnetization in the Y direction of a spin chain.

    :param absolute: Specifies whether to estimate the absolute magnetization.
    :type absolute: bool
    """

    def __init__(self, absolute=False):
        self.name = "SigmaY"
        self.symbol = "Y"
        self.absolute = absolute

    def apply(self, nn_state, samples):
        r"""Computes the magnetization along Y of each sample in the given batch of samples.

        Assumes that the computational basis that the WaveFunction was trained
        on was the Z basis.

        :param nn_state: The WaveFunction that drew the samples.
        :type nn_state: qucumber.nn_states.WaveFunctionBase
        :param samples: A batch of samples to calculate the observable on.
                        Must be using the :math:`\sigma_i = 0, 1` convention.
        :type samples: torch.Tensor
        """
        samples = samples.to(device=nn_state.device)

        # vectors of shape: (2, num_samples,)
        psis = nn_state.psi(samples)
        psi_ratio_sum = torch.zeros_like(psis)

        for i in range(samples.shape[-1]):  # sum over spin sites

            coeff = -to_pm1(samples[:, i])
            coeff = cplx.make_complex(torch.zeros_like(coeff), coeff)

            flip_spin(i, samples)  # flip the spin at site i

            # compute ratio of psi_(-i) / psi, multiply it by the appropriate
            # eigenvalue, and add it to the running sum
            psi_ratio = nn_state.psi(samples)
            psi_ratio = cplx.elementwise_division(psi_ratio, psis)
            psi_ratio = cplx.elementwise_mult(psi_ratio, coeff)
            psi_ratio_sum.add_(psi_ratio)

            flip_spin(i, samples)  # flip it back

        # take real part (imaginary part should be approximately zero)
        # and divide by number of spins
        res = cplx.real(psi_ratio_sum).div_(samples.shape[-1])
        if self.absolute:
            return res.abs_()
        else:
            return res


class SigmaZ(ObservableBase):
    r"""The :math:`\sigma_z` observable.

    Computes the magnetization in the Z direction of a spin chain.

    :param absolute: Specifies whether to estimate the absolute magnetization.
    :type absolute: bool
    """

    def __init__(self, absolute=False):
        self.name = "SigmaZ"
        self.symbol = "Z"
        self.absolute = absolute

    def apply(self, nn_state, samples):
        r"""Computes the magnetization along Z of each sample given a batch of samples.

        Assumes that the computational basis that the WaveFunction was trained
        on was the Z basis.

        :param nn_state: The WaveFunction that drew the samples.
        :type nn_state: qucumber.nn_states.WaveFunctionBase
        :param samples: A batch of samples to calculate the observable on.
                        Must be using the :math:`\sigma_i = 0, 1` convention.
        :type samples: torch.Tensor
        """
        # convert to +/- 1 convention, *after* computing the
        # mean to reduce total computations
        res = to_pm1(samples.mean(1))
        if self.absolute:
            return res.abs_()
        else:
            return res



# Copyright 2019 PIQuIL - All Rights Reserved.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from inspect import signature

from .callback import CallbackBase


class LambdaCallback(CallbackBase):
    """Class for creating simple callbacks.

    This callback is constructed using the passed functions that will be called
    at the appropriate time.

    :param on_train_start: A function to be called at the start of the training
        cycle. Must follow the same signature as
        :func:`CallbackBase.on_train_start<CallbackBase.on_train_start>`.
    :type on_train_start: callable or None

    :param on_train_end: A function to be called at the end of the training
        cycle. Must follow the same signature as
        :func:`CallbackBase.on_train_end<CallbackBase.on_train_end>`.
    :type on_train_end: callable or None

    :param on_epoch_start: A function to be called at the start of every epoch.
        Must follow the same signature as
        :func:`CallbackBase.on_epoch_start<CallbackBase.on_epoch_start>`.
    :type on_epoch_start: callable or None

    :param on_epoch_end: A function to be called at the end of every epoch.
        Must follow the same signature as
        :func:`CallbackBase.on_epoch_end<CallbackBase.on_epoch_end>`.
    :type on_epoch_end: callable or None

    :param on_batch_start: A function to be called at the start of every batch.
        Must follow the same signature as
        :func:`CallbackBase.on_batch_start<CallbackBase.on_batch_start>`.
    :type on_batch_start: callable or None

    :param on_batch_end: A function to be called at the end of every batch.
        Must follow the same signature as
        :func:`CallbackBase.on_batch_end<CallbackBase.on_batch_end>`.
    :type on_batch_end: callable or None
    """

    @staticmethod
    def _validate_function(fn, num_params, name):
        if callable(fn):
            if len(signature(fn).parameters) == num_params:
                return fn
            else:
                raise ValueError(
                    "Given function for {} must have {} arguments.".format(
                        name, num_params
                    )
                )
        elif fn is None:
            return lambda *args: None
        else:
            raise TypeError(f"{name} must be either None or a function")

    def __init__(
        self,
        on_train_start=None,
        on_train_end=None,
        on_epoch_start=None,
        on_epoch_end=None,
        on_batch_start=None,
        on_batch_end=None,
    ):
        super().__init__()
        self.on_train_start = self._validate_function(
            on_train_start, 1, "on_train_start"
        )
        self.on_train_end = self._validate_function(on_train_end, 1, "on_train_end")

        self.on_epoch_start = self._validate_function(
            on_epoch_start, 2, "on_epoch_start"
        )
        self.on_epoch_end = self._validate_function(on_epoch_end, 2, "on_epoch_end")

        self.on_batch_start = self._validate_function(
            on_batch_start, 3, "on_batch_start"
        )
        self.on_batch_end = self._validate_function(on_batch_end, 3, "on_batch_end")



# Copyright 2019 PIQuIL - All Rights Reserved.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import numpy as np

from .utils import _update_statistics


class System:
    r"""A class representing a physical system.

    It keeps track of multiple observables which it can evaluate simultaneously.

    :param \*observables: The Observables to evaluate.
    """

    def __init__(self, *observables):
        self.observables = {obs.name: obs for obs in observables}

    def statistics(self, nn_state, num_samples, num_chains=0, burn_in=1000, steps=1):
        """Estimates the expected value, variance, and the standard error of the
        observables over the distribution defined by `nn_state`.

        :param nn_state: The WaveFunction to draw samples from.
        :type nn_state: qucumber.nn_states.WaveFunctionBase
        :param num_samples: The number of samples to draw. The actual number of
                            samples drawn may be slightly higher if
                            `num_samples % num_chains != 0`.
        :type num_samples: int
        :param num_chains: The number of Markov chains to run in parallel;
                           if 0 or greater than `num_samples`, will use a
                           number of chains equal to `num_samples`. This is not
                           recommended in the case where a `num_samples` is
                           large, as this may use up all the available memory.
        :type num_chains: int
        :param burn_in: The number of Gibbs Steps to perform before recording
                        any samples.
        :type burn_in: int
        :param steps: The number of Gibbs Steps to take between each sample.
        :type steps: int
        :returns: A dictionary of dictionaries. At the top level, the keys
                  will be the names of the observables this object is keeping
                  track of. The values will be dictionaries containing the
                  (estimated) expected value (key: "mean"), variance (key:
                  "variance"), and standard error (key: "std_error") of the
                  corresponding observable.
        :rtype: dict(str, dict(str, float))
        """
        means = {name: 0.0 for name in self.observables.keys()}
        variances = {name: 0.0 for name in self.observables.keys()}
        total_samples = 0.0

        chains = None
        num_chains = min(num_chains, num_samples) if num_chains != 0 else num_samples
        num_time_steps = int(np.ceil(num_samples / num_chains))
        for i in range(num_time_steps):
            num_gibbs_steps = burn_in if i == 0 else steps

            chains = nn_state.sample(
                num_samples=num_chains,
                k=num_gibbs_steps,
                initial_state=chains,
                overwrite=True,
            )

            for obs_name, obs in self.observables.items():
                obs_stats = obs.statistics_from_samples(nn_state, chains)

                means[obs_name], variances[obs_name], _ = _update_statistics(
                    means[obs_name],
                    variances[obs_name],
                    total_samples,
                    obs_stats["mean"],
                    obs_stats["variance"],
                    num_chains,
                )

            total_samples += num_chains

        statistics = {
            obs_name: {
                "mean": means[obs_name],
                "variance": variances[obs_name],
                "std_error": np.sqrt(variances[obs_name] / total_samples),
            }
            for obs_name in self.observables.keys()
        }

        return statistics

    def statistics_from_samples(self, nn_state, samples):
        """Estimates the expected value, variance, and the standard error of the
        observables using the given samples.

        :param nn_state: The WaveFunction that drew the samples.
        :type nn_state: qucumber.nn_states.WaveFunctionBase
        :param samples: A batch of sample states to calculate the observable on.
        :type samples: torch.Tensor
        :returns: A dictionary of dictionaries. At the top level, the keys
                  will be the names of the observables this object is keeping
                  track of. The values will be dictionaries containing the
                  (estimated) expected value (key: "mean"), variance (key:
                  "variance"), and standard error (key: "std_error") of the
                  corresponding observable.
        :rtype: dict(str, dict(str, float))
        """
        statistics = {
            obs_name: obs.statistics_from_samples(nn_state, samples)
            for obs_name, obs in self.observables.items()
        }
        return statistics



# Copyright 2019 PIQuIL - All Rights Reserved.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import torch
from torch.nn.utils.convert_parameters import _check_param_device


# based on code found in PyTorch
def vector_to_grads(vec, parameters):
    r"""Convert one vector to the parameters.

    :param vec: a single vector represents the parameters of a model.
    :type vec:  torch.Tensor
    :param parameters: an iterator of Tensors that are the parameters of a
                       model.
    :type parameters: list[torch.Tensor]
    """
    # Ensure vec of type Tensor
    if not isinstance(vec, torch.Tensor):
        raise TypeError(
            "expected torch.Tensor, but got: {}".format(torch.typename(vec))
        )
    # Flag for the device where the parameter is located
    param_device = None

    # Pointer for slicing the vector for each parameter gradient
    pointer = 0
    for param in parameters:
        # Ensure the parameters are located in the same device
        param_device = _check_param_device(param, param_device)

        # The length of the parameter
        num_param = param.numel()

        # Slice the vector, reshape it, and replace the gradient data of
        # the parameter
        param.grad = vec[pointer : pointer + num_param].view(param.size()).data

        # Increment the pointer
        pointer += num_param



# Copyright 2019 PIQuIL - All Rights Reserved.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


class CallbackBase:
    """Base class for callbacks."""

    def on_train_start(self, nn_state):
        """Called at the start of the training cycle.

        :param nn_state: The WaveFunction being trained.
        :type nn_state: qucumber.nn_states.WaveFunctionBase
        """
        pass

    def on_train_end(self, nn_state):
        """Called at the end of the training cycle.

        :param nn_state: The WaveFunction being trained.
        :type nn_state: qucumber.nn_states.WaveFunctionBase
        """
        pass

    def on_epoch_start(self, nn_state, epoch):
        """Called at the start of each epoch.

        :param nn_state: The WaveFunction being trained.
        :type nn_state: qucumber.nn_states.WaveFunctionBase
        :param epoch: The current epoch.
        :type epoch: int
        """
        pass

    def on_epoch_end(self, nn_state, epoch):
        """Called at the end of each epoch.

        :param nn_state: The WaveFunction being trained.
        :type nn_state: qucumber.nn_states.WaveFunctionBase
        :param epoch: The current epoch.
        :type epoch: int
        """
        pass

    def on_batch_start(self, nn_state, epoch, batch):
        """Called at the start of each batch.

        :param nn_state: The WaveFunction being trained.
        :type nn_state: qucumber.nn_states.WaveFunctionBase
        :param epoch: The current epoch.
        :type epoch: int
        :param batch: The current batch index.
        :type batch: int
        """
        pass

    def on_batch_end(self, nn_state, epoch, batch):
        """Called at the end of each batch.

        :param nn_state: The WaveFunction being trained.
        :type nn_state: qucumber.nn_states.WaveFunctionBase
        :param epoch: The current epoch.
        :type epoch: int
        :param batch: The current batch index.
        :type batch: int
        """
        pass



# Copyright 2019 PIQuIL - All Rights Reserved.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import torch
from torch.distributions.utils import probs_to_logits

import qucumber.utils.cplx as cplx
import qucumber.utils.unitaries as unitaries


def fidelity(nn_state, target_psi, space, **kwargs):
    r"""Calculates the square of the overlap (fidelity) between the reconstructed
    wavefunction and the true wavefunction (both in the computational basis).

    :param nn_state: The neural network state (i.e. complex wavefunction or
                     positive wavefunction).
    :type nn_state: WaveFunction
    :param target_psi: The true wavefunction of the system.
    :type target_psi: torch.Tensor
    :param space: The hilbert space of the system.
    :type space: torch.Tensor
    :param \**kwargs: Extra keyword arguments that may be passed. Will be ignored.

    :returns: The fidelity.
    :rtype: float
    """
    Z = nn_state.compute_normalization(space)
    target_psi = target_psi.to(nn_state.device)
    psi = nn_state.psi(space) / Z.sqrt()
    F = cplx.inner_prod(target_psi, psi)
    return cplx.absolute_value(F).pow_(2).item()


def rotate_psi(nn_state, basis, space, unitaries, psi=None):
    r"""A function that rotates the reconstructed wavefunction to a different
    basis.

    :param nn_state: The neural network state (i.e. complex wavefunction or
                     positive wavefunction).
    :type nn_state: WaveFunction
    :param basis: The basis to rotate the wavefunction to.
    :type basis: str
    :param space: The hilbert space of the system.
    :type space: torch.Tensor
    :param unitaries: A dictionary of (2x2) unitary operators.
    :type unitaries: dict
    :param psi: A wavefunction that the user can input to override the neural
                network state's wavefunction.
    :type psi: torch.Tensor

    :returns: A wavefunction in a new basis.
    :rtype: torch.Tensor
    """
    psi = (
        nn_state.psi(space)
        if psi is None
        else psi.to(dtype=torch.double, device=nn_state.device)
    )

    unitaries = {k: v.to(device=nn_state.device) for k, v in unitaries.items()}
    us = [unitaries[b] for b in basis]
    n_u = [u.size()[0] for u in us]

    l = 1  # noqa: E741
    for n in n_u:
        l *= n  # noqa: E741

    r = 1

    psi_r = psi.clone()
    for s in range(len(n_u))[::-1]:
        l //= n_u[s]  # noqa: E741
        m = us[s]

        for k in range(l):
            for i in range(r):
                slc = slice(k * n_u[s] * r + i, (k + 1) * n_u[s] * r + i, r)
                U = psi_r[:, slc]
                psi_r[:, slc] = cplx.matmul(m, U)
        r *= n_u[s]

    return psi_r


def NLL(nn_state, samples, space, train_bases=None, **kwargs):
    r"""A function for calculating the negative log-likelihood.

    :param nn_state: The neural network state (i.e. complex wavefunction or
                     positive wavefunction).
    :type nn_state: WaveFunction
    :param samples: Samples to compute the NLL on.
    :type samples: torch.Tensor
    :param space: The hilbert space of the system.
    :type space: torch.Tensor
    :param train_bases: An array of bases where measurements were taken.
    :type train_bases: np.array(dtype=str)
    :param \**kwargs: Extra keyword arguments that may be passed. Will be ignored.

    :returns: The Negative Log-Likelihood.
    :rtype: float
    """
    psi_r = torch.zeros(
        2, 1 << nn_state.num_visible, dtype=torch.double, device=nn_state.device
    )
    NLL = 0.0
    unitary_dict = unitaries.create_dict()
    Z = nn_state.compute_normalization(space)
    eps = 0.000001
    if train_bases is None:
        for i in range(len(samples)):
            NLL -= (cplx.norm_sqr(nn_state.psi(samples[i])) + eps).log()
            NLL += Z.log()
    else:
        for i in range(len(samples)):
            # Check whether the sample was measured the reference basis
            is_reference_basis = True
            # b_ID = 0
            for j in range(nn_state.num_visible):
                if train_bases[i][j] != "Z":
                    is_reference_basis = False
                    break
            if is_reference_basis is True:
                NLL -= (cplx.norm_sqr(nn_state.psi(samples[i])) + eps).log()
                NLL += Z.log()
            else:
                psi_r = rotate_psi(nn_state, train_bases[i], space, unitary_dict)
                # Get the index value of the sample state
                ind = 0
                for j in range(nn_state.num_visible):
                    if samples[i, nn_state.num_visible - j - 1] == 1:
                        ind += pow(2, j)
                NLL -= cplx.norm_sqr(psi_r[:, ind]).log().item()
                NLL += Z.log()
    return (NLL / float(len(samples))).item()


def KL(nn_state, target_psi, space, bases=None, **kwargs):
    r"""A function for calculating the total KL divergence.

    :param nn_state: The neural network state (i.e. complex wavefunction or
                     positive wavefunction).
    :type nn_state: WaveFunction
    :param target_psi: The true wavefunction of the system.
    :type target_psi: torch.Tensor
    :param space: The hilbert space of the system.
    :type space: torch.Tensor
    :param bases: An array of unique bases.
    :type bases: np.array(dtype=str)
    :param \**kwargs: Extra keyword arguments that may be passed. Will be ignored.

    :returns: The KL divergence.
    :rtype: float
    """
    psi_r = torch.zeros(
        2, 1 << nn_state.num_visible, dtype=torch.double, device=nn_state.device
    )
    KL = 0.0
    target_psi = target_psi.to(nn_state.device)
    Z = nn_state.compute_normalization(space)
    if bases is None:
        target_probs = cplx.absolute_value(target_psi) ** 2
        nn_probs = nn_state.probability(space, Z)

        KL += torch.sum(target_probs * probs_to_logits(target_probs))
        KL -= torch.sum(target_probs * probs_to_logits(nn_probs))
    else:
        unitary_dict = nn_state.unitary_dict
        for basis in bases:
            psi_r = rotate_psi(nn_state, basis, space, unitary_dict)
            target_psi_r = rotate_psi(nn_state, basis, space, unitary_dict, target_psi)

            probs_r = (cplx.absolute_value(psi_r) ** 2) / Z
            target_probs_r = cplx.absolute_value(target_psi_r) ** 2

            KL += torch.sum(target_probs_r * probs_to_logits(target_probs_r))
            KL -= torch.sum(target_probs_r * probs_to_logits(probs_r))
        KL /= float(len(bases))

    return KL.item()
